
#include <stdint.h>
#include <pic.h>
#include <xc.h>
#include "PIC16F184xx_ADC.h"
#include "main.h"

static uint16_t g_adcBuf[ADC_BUF_SIZE];
void ADC_Push(uint16_t d)
{
	static uint8_t i;
	g_adcBuf[i++] = d;
	if(i>=ADC_BUF_SIZE) i=0;
}

void ADC_Init(void)
{
	// FVR
	FVRCONbits.ADFVR = 2; // ADC FVR Buffer is x2 (2.048V)
	FVRCONbits.FVREN = 1;
	
	
	// Precharge
	// if PRE != 0, ADCON1.PPOL selects action during 1st precharge stage
//	ADPRE = 0; // 0us
	// Acquisition time
//	ADACQ = 0; // 0us
	
	ADCON0bits.ON = 1; // ADC ON
	// CONT = 0
	ADCON0bits.CS = 0;
	ADCON0bits.FM = 1; // right-justified. this bit does not meet with datasheet
	// CS = 0 (clock uses FOSC/N)
	//FRM = 0
	ADCON2bits.MD = 0; // Basic Mode
	
	// Trigger Source: Timer 2
	ADACT = 0x04; // Auto Trigger = TMR2_postscaled
	ADPCH = 0x05; // CH+
	
	ADREF = 0b00000000; // VREF-: AVSS, VREF+: VDD
//	ADREF = 0b00000011; // VREF-: AVSS, VREF+: FVR
	
	ADCLK = 31;// FOSC/(2*CS+2). FOSC = 32MHz/64 = 500kHz Clock
	// TAD = typ. 2us
	
//	ADCPCON0 = 0x00; // charge pump
	
	// Interrupt
	PIE1bits.ADIE = 1; // interrupt enable
}

uint16_t ADC_OneShot(uint8_t channel)
{
	ADPCH = channel;	
	// Start
	ADCON0bits.GO = 1;
	while(ADCON0bits.GO);
	
	return ADRES; // Resulat
}

void ADC_Start(uint8_t channel)
{
	ADPCH = channel;
	// Start
	ADCON0bits.GO = 1;
}
uint16_t g_refValue;
void ADC_SetDCDCVRefValue(uint8_t biasVoltage)
{
	g_refValue = (1000U*(uint32_t)biasVoltage + 7000U) * 
					ADC_FS * DCDC_FB_R2 / (DCDC_FB_R2 + DCDC_FB_R1) / ADC_VREF;
}
void ADC_IRQ(void) // Call from general IRQ routine
{
	uint16_t res = ADRES;
	ADC_Push(ADRES);
	// set next channel
	
	// DC-DC Control
#if (DCDC_CONTROL_MODE == 1) // Simple comparator
	{
		if(g_refValue > res){
			PWM6DC = PWM6_DUTY_MAX << 6;
		}else{
			PWM6DC = 0;
		}
	}
#else
	{
#define PWM_DIFF_MAX 16
		static int16_t pwmValue = 0;
		int16_t diff = ((int16_t)g_refValue - (int16_t)res) ;
		diff *= 4;
		if(diff > PWM_DIFF_MAX) diff = PWM_DIFF_MAX;
		if(diff < -PWM_DIFF_MAX) diff = -PWM_DIFF_MAX;
		if(diff > 0) {
			if((pwmValue + diff) < PWM6_DUTY_MAX) pwmValue += diff;
			else pwmValue = PWM6_DUTY_MAX-1;
		}else {
			if(pwmValue >= -diff) pwmValue += diff;
			else pwmValue = 0;
		}
		PWM6DC = pwmValue << 6; // duty cycle (shifted by 6-bits)
	}
#endif
	// if start new conversion here, it's too fast
}
uint16_t ADC_GetCurrentSum()		
{
	uint8_t i;
	uint16_t ret = 0;
	for(i=0; i<ADC_BUF_SIZE; i++) ret += g_adcBuf[i];
	return ret;
}
