/*
 * File:   main.c
 * Author: Owner
 *
 * Created on 2020/11/03, 16:39
 */


#include <stdint.h>
#include <pic.h>
#include <xc.h>
#include "main.h"
#include "PIC16F184xx_ADC.h"
#include "SwitchProc.h"

const uint8_t g_fontTable[12] = { // 0-9, n, u
0b00111111, 0b00000110, 0b01011011, 0b01001111, 0b01100110,
0b01101101, 0b01111101, 0b00100111, 0b01111111, 0b01101111,
0b01010100, 0b00011100};
//0b01010100, 0b01110010};
uint8_t g_digitData[8];
volatile uint8_t g_swPortState;
volatile uint8_t g_100msFlag;
volatile uint8_t g_timeoutFlag;
volatile uint8_t g_outputEnable; // 0: disable, 1: enable, 2: auto mode

/*
#define CAP_PERIOD_BUF_SIZE 48
static uint32_t g_capPeriodBuf[CAP_PERIOD_BUF_SIZE];
void CapPeriod_Push(uint32_t d)
{
	static uint8_t i;
	g_capPeriodBuf[i++] = d;
	if(i>=CAP_PERIOD_BUF_SIZE) i=0;
}
uint32_t CapPeriod_GetCurrentSum()		
{
	uint8_t i;
	uint32_t ret = 0;
	for(i=0; i<CAP_PERIOD_BUF_SIZE; i++) ret += g_capPeriodBuf[i];
	return ret;
}
*/


static void SREGSend(uint16_t d)
{
	uint8_t i;
	for(i=0; i<16; i++)
	{
		LATCbits.LATC4 = 0;
		if(d & 0x8000){
			LATCbits.LATC5 = 0;
		}else{
			LATCbits.LATC5 = 1;
		}
		d = d << 1;
		LATCbits.LATC4 = 1;
	}
	LATAbits.LATA1 = 0;
	asm("nop");
	LATAbits.LATA1 = 1;
}
static void Hex2BCD(uint16_t d, uint8_t* pBCD)
{
	// output pBCD[4](MSD) - pBCD[0](LSD)
//	pBCD[4] = d/10000;
//	d -= pBCD[4] * 10000U;
	pBCD[3] = d/1000;
	d -= pBCD[3] * 1000U;
	pBCD[2] = d/100;
	d -= pBCD[2] * 100U;
	pBCD[1] = d/10;
	pBCD[0] = d - pBCD[1]*10U;
}
// voltage: 0-31, 32, 34,..., 62
// output voltage: setting voltage * 1.024 / 32 V
static void DAC_SetVoltage(uint8_t voltage)
{
	if(voltage >= 32) {
		DAC1CON1 = voltage / 2;
		FVRCONbits.CDAFVR = 2; // x2
	}else{
		if(FVRCONbits.CDAFVR == 2) {
			FVRCONbits.CDAFVR = 1; // x1
		}
		DAC1CON1 = voltage;
	}
}
uint8_t UART_SendByte(uint8_t ch)
{
	while(TX1STAbits.TRMT == 0);
	TX1REG = ch;
	return 0;
}
uint8_t UART_SendText(const char* p)
{
	uint8_t l=128;
	while(l--){
		if(*p == '\0') break;
		if(UART_SendByte(*p)) return -1;
		p++;
	}
	return 0;
}
#define UART_SendInitMessage() UART_SendText(UART_INIT_MESSAGE) 

void UART_SendResult(const uint8_t* bcd, uint8_t capRange)
{
	if(bcd[5]) UART_SendByte('0' + bcd[5]);
	UART_SendByte('0' + bcd[4]);
	UART_SendByte(UART_SEPARATOR);
	switch(capRange){
	case 0: // 0.00n
//		UART_SendByte('0' + bcd[3]);
		UART_SendByte('0' + bcd[2]);
		UART_SendByte('.');
		UART_SendByte('0' + bcd[1]);
		UART_SendByte('0' + bcd[0]);
		break;
	case 1: // 00.0n
//		UART_SendByte('0' + bcd[3]);
		UART_SendByte('0' + bcd[2]);
		UART_SendByte('0' + bcd[1]);
		UART_SendByte('.');
		UART_SendByte('0' + bcd[0]);
		break;
	case 2: // 000n
//		UART_SendByte('0' + bcd[3]);
		UART_SendByte('0' + bcd[2]);
		UART_SendByte('0' + bcd[1]);
		UART_SendByte('0' + bcd[0]);
		break;
	case 3: // 0.00u
//		UART_SendByte('0' + bcd[3]);
		UART_SendByte('0' + bcd[2]);
		UART_SendByte('0' + bcd[1]);
		UART_SendByte('0' + bcd[0]);
		UART_SendByte('0');
		break;
	case 4: // 00.0u
//		UART_SendByte('0' + bcd[3]);
		UART_SendByte('0' + bcd[2]);
		UART_SendByte('0' + bcd[1]);
		UART_SendByte('0' + bcd[0]);
		UART_SendByte('0');
		UART_SendByte('0');
		break;
	case 5: // 000u
//		UART_SendByte('0' + bcd[3]);
		UART_SendByte('0' + bcd[2]);
		UART_SendByte('0' + bcd[1]);
		UART_SendByte('0' + bcd[0]);
		UART_SendByte('0');
		UART_SendByte('0');
		UART_SendByte('0');
		break;
	}
	UART_SendByte('\r');
	UART_SendByte('\n');
}

void CapacitorOscSetRange(uint8_t range)
{
	if(range){ // high range
		RA2PPS = 0x12; // C2OUT (2)
		TRISAbits.TRISA2 = 0; // Set as output 
	}else{
		TRISAbits.TRISA2 = 1; // Set as input
		RA2PPS = 0;
	}
}

// pCapValueBCD: uint8_t[5]
// pCapRange: 0 -> xx.xx nF 1 -> xxx.x nF 2-> x.xxx uF 3 -> xx.xx uF 4 -> OL
void ConvertToCapacitance(uint32_t cnt, uint8_t range, uint8_t* pCapValueBCD, uint8_t* pCapRange)
{
	/*
	static uint8_t lastRange;
	if(range == 1){// High Range
		RA2PPS = 0x12; // C2OUT (2)
		TRISAbits.TRISA2 = 0; // Set as output 
	}else{
		TRISAbits.TRISA2 = 1; // Set as input 
	}*/

	// Easy version
	// If constant current given, capacitance Cx = It/dV
//	float cap = VCC/2/RSRC1*cnt/COUNT_CLOCK/(VCC/VCC_PER_VREF*2)
	uint32_t cap;
	if(range){
		cap = (uint32_t)(cnt*4.98/* /0.201 */ /COUNT_CLOCK*500/RSRC2); // pF
//		cap = cnt*VCC_PER_VREF/COUNT_CLOCK*1000/RSRC2/4; // pF
	}else{
		cap = (uint32_t)(cnt*4.98/* /0.201 */ /COUNT_CLOCK*500/RSRC1); // pF
//		cap = cnt*VCC_PER_VREF/COUNT_CLOCK*1000/RSRC1/4; // pF
	}
#if 1
	for(*pCapRange = 0; *pCapRange < 6; (*pCapRange) ++)
	{
		cap = cap / 10;
		if(cap < 999){
			Hex2BCD(cap, pCapValueBCD); 
			break;
		}
	}
	if(*pCapRange == 6)
	{
		pCapValueBCD[0] = 9;
		pCapValueBCD[1] = 9;
		pCapValueBCD[2] = 9;
//		pCapValueBCD[3] = 9;
	}
#else
	// range
	if(cap > 999999999) { // xxxx.uF
		pCapValueBCD[0] = 9;
		pCapValueBCD[1] = 9;
		pCapValueBCD[2] = 9;
		pCapValueBCD[3] = 9;
		*pCapRange = 5;
	}
	else if(cap > 99999999) { // xxx.xuF
		Hex2BCD(cap / 100000, pCapValueBCD); 
		*pCapRange = 4;
	}else if(cap > 9999999){ // xx.xx uF
		Hex2BCD(cap / 10000, pCapValueBCD); 
		*pCapRange = 3;
	}else if(cap > 999999){ // x.xxx uF
		Hex2BCD(cap / 1000, pCapValueBCD); 
		*pCapRange = 2;
	}else if(cap > 99999){ // xxx.x nF
		Hex2BCD(cap / 100, pCapValueBCD); 
		*pCapRange = 1;
	}else{ // xx.xx nF
		Hex2BCD(cap / 10, pCapValueBCD); 
		*pCapRange = 0;
	}
#endif
}

// interrupt (18424 supports only 1 level interrupt)
void __interrupt () SYS_InterruptHigh(void)
{
	if(PIR0bits.TMR0IF) // every 1ms (this takes some processing time)
	{
		static uint8_t localCnt;
		static uint8_t ledBlinkFlag;
		localCnt ++;
		PIR0bits.TMR0IF = 0;
		
		// get switch state
		uint8_t swPortState = PORTC & 0x30U; // RC4 and RC5
		if(PORTAbits.RA3) swPortState |= 0x01;
		
		TRISCbits.TRISC5 = 0; // output
		TRISCbits.TRISC4 = 0; // output
		
		uint8_t digit = (localCnt & 0xF) / 2;
		uint16_t d = ((uint16_t)g_digitData[digit]);
		if(localCnt & 1) d |= (0x0100U << digit);
		if((g_outputEnable == 1) || ((g_outputEnable == 2) && (ledBlinkFlag & 2))){
			d |= 0x1000;
		}else{
			d &= ~0x1000;
		}
		SREGSend(d);
		
		// set as input for switch
		TRISCbits.TRISC5 = 1; // input
		TRISCbits.TRISC4 = 1; // input
		
		if(localCnt >= 128){
			g_100msFlag = 1;
			localCnt = 0;
			ledBlinkFlag ++;
			g_timeoutFlag ++;
			
			GetSwitchState(swPortState, (uint8_t*)&g_swPortState);
		}
	}
	if(PIR1bits.ADIF) // ADC Complete
	{
		PIR1bits.ADIF = 0;
		ADC_IRQ();
	}
}
static void InitPort(void)
{
	// Port A
	TRISA  = 0b11101101; // 0: DAC_OUT, 1: LCK, 2: RANGE, 3: MCLR, 4: UART TX, 5: ADC Input 
	LATA   = 0b00000010; // 
	ANSELA = 0b00100000; // 0: digital pin
//	ODCONA = 0;
//	RA5PPS = 5; // default value (see p.221)
//	SLRCONAbits.SLRA2 = 0; // Full
	
	// Port C
	TRISC  = 0b11001101; // 0: Comp In+ 1: CLC_OUT, 2: Comp IN-, 3: PWM, 4: DCK, 5: DATA
	ANSELC = 0b00000101;
//	ODCONC = 0;
//	SLRCONCbits.SLRC1 = 0; // Full
	
	WPUA   = 0b00001000;
	WPUC   = 0b00110000;
	
//	SREGSend(0);

	// Timer (1ms = 8M/8/4/256)
	T0CON0bits.T016BIT = 0; // 8bit
	T0CON0bits.T0OUTPS = 0;
	
	T0CON1bits.ASYNC = 1; // Asynchronous (to wake up in sleep mode)
	T0CON1bits.T0CS = 2; // FOSC/4
//	if(OPERATE_IN_1M){
//		T0CON1bits.T0CKPS = 0; // 1/1
//	}else{
		T0CON1bits.T0CKPS = 5; // 1/32
//	}
//	TMR0L = 0;
	T0CON0bits.T0EN = 1;
	
	PIE0bits.TMR0IE = 1;
	CPUDOZEbits.IDLEN = 1;
	
	// Timer 2 (ADC Trigger = 20kHz)
	T2PR  = 50;
	T2CON = 0b00100000; // CKPS = 1/4
	T2CLKCON = 1; // FOSC/4
	T2CONbits.ON = 1;
	
	// Timer 4 (22KHz = 32M/1454)
//	PWM6CON = 0; // (default value)
	T4PR = PWM6_DIV;
	PWM6DC = 0; // duty cycle (shift by 6-bits)
	T4CON = 0b00010000; // CKPS = 1/1, OUTPS = 1/1
//	T4HLT = 0; // Period pulse (default value)
	T4CLKCON = 1; // FOSC/4 (not FOSC)
	T4CONbits.ON = 1;
	PIR4bits.TMR4IF = 0; //necessary for PWM
	
#if 0
	// CLC1: XOR (IN1 = COMP, IN2 = CLC2_OUT). lxcg2,4 = L
	
	// CLC2: 2-input D-FF (D-IN = CLC1_OUT, CLK = FCLK). l2 = D1, l4 = D2, l1 = CLK, l3 = L
	CLC2CON = 0b00000101;
	CLC2POL = 0b10001010; // invert output, l2,l4: invert (NOR)
	CLC2SEL0 = 28;// D1S (input source = COMP2)
	CLC2SEL1 = 33;// D1S (input source = CLC2_out)
	CLC2SEL2 = 0;// D1S
	CLC2SEL3 = 4;// FOSC
	CLC2GLS0 = 0b10000000; // Select G4
	CLC2GLS1 = 0b00000101; // Select /G1 and /G2
	CLC2GLS2 = 0;// G3D4T/N - G3D1T/N
	CLC2GLS3 = 0b00001010;// Select G1 and G2
	// debug
//	CLC2GLS1 = 0b00000001; // Select /G1
//	CLC2GLS3 = 0b00000001; // Select /G1
	CLC2CONbits.EN = 1;
	
	// CLC3: J-K FF (J = COMP and CLC2_OUT, K = /COMP and CLC2_OUT, CLK = FCLK). 
	
	// CLC4: Thru (Transparent Latch). l2 = D, l3 = LE (H), l1 = L, l4 = L
	CLC4CON = 0b00000100;
	CLC4POL = 0b00000000;
	CLC4SEL0 = 28;// D1S (input source = COMP2)
	CLC4SEL1 = 33;// D1S (input source = CLC2_out)
	CLC4SEL2 = 0;// D1S
	CLC4SEL3 = 4;// FOSC
	CLC4GLS0 = 0b10000000; 
	CLC4GLS1 = 0b00000010; // Select D1
	CLC4GLS2 = 0;
	CLC4GLS3 = 0;
	CLC4CONbits.EN = 1;
#endif
	
	// (debug) output from RA4
//	ANSELAbits.ANSA4 = 0;
//	TRISAbits.TRISA4 = 0;
//	RA4PPS = 0x04; // CLC4
	RA4PPS = 0x0F; // TX1 (UART)
	
	// PWM6 (also must disable TRIS bits of PWMx pin)
	CCPTMRS1bits.P6TSEL = 2;// Select TMR4
	PWM6CON = 0x80;
	
	// Configure PWM6 pin
	RC3PPS = 0x0D; // PWM6OUT
	TRISCbits.TRISC3 = 0; // output PWM pin
	
	// Analog Comparator
	CM2CON0 = 0b10000000; // enable comp 2, inverted output
//	CM2NCH  = 3; // C2IN3- (RC3)
	CM2NCH  = 2; // C2IN2- (RC2)
	CM2PCH  = 0; // C2IN0+ (RC0)
	// Assign output to RC1
	PPSLOCK = 0;
	RC1PPS = 0x12; // C2OUT (1)
	CapacitorOscSetRange(0);

//	LATCbits.LATC1 = 1;//test
	
	// CLC (RS-FF, Set with ACOMP-, Clear with ACOMP+)
	// lcxg1 | lxcg2 = S
	// lcxg3 | lcxg4 = R
	// lcxq = Q
/*
	CLC1POL = 0;
	// Selector
	CLC1SEL0 = 28; // D1S = C2_out
	CLC1SEL1 = 0;
	CLC1SEL2 = 0;
	CLC1SEL3 = 0;
	// Gate
	CLC1GLS0 = 0; // DyT/DyN
	CLC1GLS1 = 0;
	CLC1GLS2 = 0;
	CLC1GLS3 = 0;
	CLC1CON = 0b10000011;
*/
	// SMT
	SMT1CON0 = 0b10000000;
	SMT1CON1 = 0b00000010; // One shot, MODE = Period and duty-cycle, GO=1
	SMT1CLK = 1; // FOSC
	SMT1WIN = 0; // Window: PPS
	SMT1SIG = 0x10; // Signal = C2OUT
	// 32M / 16,777,216 = 520ms
	
	// UART (19.2k)
	TX1STA = 0b00100000; // TXEN = 1
	BAUD1CON = 0b00000000;
	SP1BRG = 25;
	RC1STA = 0b10000000; // SPEN = 1
	
	
	// ADC
	ADC_Init();
	
	// DAC
	FVRCONbits.CDAFVR = 1; // FVR = x1
	FVRCONbits.FVREN = 1;
	DAC1CON0 = 0b10101000; // DAC+ = FVR, DAC- = AVSS
	
	
	// Interrupt
	INTCONbits.PEIE = 1;
	INTCONbits.GIE = 1;
	
	uint8_t i;
	for(i=0; i<8; i++) g_digitData[i] = g_fontTable[i];
	
//	__delay_ms(100);
	UART_SendInitMessage();
}

void main(void) 
{
	uint32_t period0, period1; // period accumlator
	uint8_t curRange = 0;
	uint16_t cap_offset = 0;
	// State machine states
#define PHASE_NORMAL 0
#define PHASE_START 1
#define PHASE_CHANGE_VOLTAGE 12
	uint8_t autoModePhaseCnt = 0; // [1] -> [2...12] -> [2] / [0]
	g_outputEnable = 0;
	InitPort();
	
	uint8_t biasVoltage = 0, biasVoltage_restore = 0;
	ADC_SetDCDCVRefValue(0);
	DAC_SetVoltage(2);
	
	while(1)
	{
		uint8_t swState = g_swPortState;
		g_swPortState = 0;
//		g_digitData[0] = (swState & 0x01)? g_fontTable[1]: g_fontTable[0];
//		g_digitData[1] = (swState & 0x10)? g_fontTable[1]: g_fontTable[0];
//		g_digitData[2] = (swState & 0x20)? g_fontTable[1]: g_fontTable[0];
		// switch procedure
		if(autoModePhaseCnt)
		{
			uint8_t restoreToNormalMode = 0;
			// Auto mode
			// State
			if(autoModePhaseCnt == PHASE_START)
			{
				UART_SendText("--Start\r\n");
				g_outputEnable = 2;
				autoModePhaseCnt ++;
			}
			else if(autoModePhaseCnt == PHASE_CHANGE_VOLTAGE)
			{
				if(biasVoltage == biasVoltage_restore)
				{
					restoreToNormalMode = 1;
				}
				else
				{
					if(biasVoltage <= 29) {
						biasVoltage ++;
					}else if(biasVoltage < VBIAS_MAX-1) {
						biasVoltage += 2;
					}
					ADC_SetDCDCVRefValue(biasVoltage); // VBIAS + 2 + 3V
					DAC_SetVoltage(biasVoltage + 2);
					autoModePhaseCnt = PHASE_START + 1;
				}
			}
			else
			{
				autoModePhaseCnt ++;
			}
			
			// Switch
			if(swState != SW_UNPUSHED)
			{
				if((swState & 0x0F) == SW_SW1){
					restoreToNormalMode = 1;
				}
			}
			
			if(restoreToNormalMode)
			{
				// restore to normal mode
				autoModePhaseCnt = PHASE_NORMAL;
				g_outputEnable = 0;
				biasVoltage = biasVoltage_restore;
				ADC_SetDCDCVRefValue(0); // VBIAS + 2 + 3V
				DAC_SetVoltage(0 + 2);
			}
		}
		else
		{
			// Normal mode
			if(swState != SW_UNPUSHED)
			{
				if((swState & 0x0F) == SW_SW2){
					if(biasVoltage >= 32) {
						biasVoltage -= 2;
					}else if(biasVoltage) {
						biasVoltage --;
					}
				}
				if((swState & 0x0F) == SW_SW3){
					if(biasVoltage <= 29) {
						biasVoltage ++;
					}else if(biasVoltage < VBIAS_MAX-1) {
						biasVoltage += 2;
					}
				}
				if((swState & 0x0F) == SW_SW1){
					if(swState == (SW_SW1 | SW_LONG)){
						if(g_outputEnable == 0){ // if current mode = disable
							// Start auto mode
							autoModePhaseCnt = PHASE_START;
							biasVoltage_restore = biasVoltage;
							biasVoltage = 0;
						}
					}else if(swState == SW_SW1){
						g_outputEnable ^= 1;
					}
				}

				if(g_outputEnable){
					ADC_SetDCDCVRefValue(biasVoltage); // VBIAS + 2 + 3V
					DAC_SetVoltage(biasVoltage + 2);
				}else{
					ADC_SetDCDCVRefValue(0); // VBIAS + 2 + 3V
					DAC_SetVoltage(0 + 2);
				}
			}
		}
		
		// repeat count acquision
		uint16_t accumCount = 0;
		period0 = period1 = 0;
		uint8_t timeOutFlag = 0;

		// ADC
		uint16_t adcResult = ADC_GetCurrentSum()/ADC_BUF_SIZE; //ADC_OneShot(0x3E); // FVR1 (1.024V)
		// ADC value -> voltage (x10mV)
		adcResult = ((uint32_t)adcResult) * ADC_VREF / ADC_FS * (DCDC_FB_R1+DCDC_FB_R2)/ DCDC_FB_R2 / 10;
		
		// Ileak can be cancelled by using harmonic avarage of duration + and -
		uint32_t lastCount = 0;
		g_timeoutFlag = 0;
		do
		{
			// SMT start
			SMT1CON1bits.GO = 1;
			// wait for SMT
			while(SMT1STATbits.TS)
			{
				if(g_timeoutFlag > 5){ // timed out & range +
					timeOutFlag = 1;
					SMT1CON1bits.GO = 0;
					break;
				}
			}
			uint32_t df = SMT1CPR; // tH + tL
			uint32_t dh = SMT1CPW; // tH
			if(df > lastCount*3/4){ // do not count pseudo-value
				period0 += df;
				period1 += dh;
				lastCount = df;
				accumCount ++;
			}
		}
		while((g_100msFlag == 0) && (period0 < 0xFF000000));
		
		if(accumCount) {
			period0 /= accumCount; // tL + tH
			period1 /= accumCount; // tH
#if CAPACITOR_HARMONICS_AVARANGE
			// Correction
			// tcor = 1/(1/tH + 1/tL))
			if(period1 && (period0 > period1)){
				uint32_t a = period0 - period1;
				uint32_t b = period1;
				period0 = 4*(0x80000000U/(0x80000000U/a + 0x80000000U/b));
			}
#endif
		}
		g_100msFlag = 0;
		
		if(autoModePhaseCnt == PHASE_NORMAL) // do not change range during auto mode
		{
			if(timeOutFlag){
				// range up
				CapacitorOscSetRange(1);
				curRange = 1;
				continue;
			}
			// range down
			if((curRange == 1) && (period0 < 0x00000FFF)){
				// range down
				CapacitorOscSetRange(0);
				curRange = 0;
				continue;
			}
		}
		
		{
			uint8_t bcd[8];
			uint8_t capRange = 0;
			// offset adjust
			if(curRange == 0){
				
				if(cap_offset == 0)
				{
					if(period0 < 0x10000) cap_offset = period0;
				}
				else
				{
					period0 = (period0 > cap_offset)? period0 - cap_offset: 0;
				}
			}
			ConvertToCapacitance(period0, curRange, bcd, &capRange);
			Hex2BCD(biasVoltage, &bcd[4]);
//			Hex2BCD(period[0] >> 8, bcd);
			if(1)
			{
				g_digitData[0] = g_fontTable[bcd[2]];
				g_digitData[1] = g_fontTable[bcd[1]];
				g_digitData[2] = g_fontTable[bcd[0]];
				if(capRange == 6){ // OL
				}else if((capRange == 2) || (capRange == 5)){
					g_digitData[2] |= 0x80;
				}else if((capRange == 0) || (capRange == 3)){
					g_digitData[0] |= 0x80;
				}else{ // 0,3
					g_digitData[1] |= 0x80;
				}
				if(capRange <= 2){
					g_digitData[3] = g_fontTable[10]; // n
				}else{
					g_digitData[3] = g_fontTable[11]; // u
				}
				if(autoModePhaseCnt == PHASE_CHANGE_VOLTAGE){
					UART_SendResult(bcd, capRange);
				}
			}else{
			
				// debug
				static uint8_t displayUpdateCnt;
				if(displayUpdateCnt++ == 20)
				{
					displayUpdateCnt = 0;
					Hex2BCD(adcResult, bcd);
					g_digitData[0] = g_fontTable[bcd[3]];
					g_digitData[1] = g_fontTable[bcd[2]];
					g_digitData[2] = g_fontTable[bcd[1]];
					g_digitData[3] = g_fontTable[bcd[0]];
				}
			}
			
			g_digitData[5] = bcd[5]? g_fontTable[bcd[5]]: 0;
			g_digitData[6] = g_fontTable[bcd[4]];
		}
		
	}
	return;
}
